/***** kerror.c ***************************************************
 * Description: Compute pairwise alignment using "exclusion
 *              algorithm".
 * Reference: Baeza-Yates, R. A. and Perlberg, C. H. (1992). Fast 
 *            and practical approximate string matching. Proc. 3rd 
 *            Symp. on Combinatorial Pattern matching. Springer 
 *            Lecture Notes in Computer Science, p. 185-192.
 *
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Tue Jan 18 08:16:52 2005.
 *****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <assert.h>
#include "stringUtil.h"
#include "interface.h"
#include "alignment.h"
#include "ktree.h"
#include "kerror.h"
#include "eprintf.h"

void checking(KTreeResult *ktr, Sequence *query, Sequence *subject, SubstitutionMatrix *sm, char *strand, Args args);

int main(int argc, char *argv[]){
  Args args;                   /* arguments */
  char *version;               /* program version */
  Sequence *query, *subject;   /* query & subject sequences */
  Sequence *rev;
  KNode *root;
  FILE *fp, *fpq, *fps;
  Sequence *patterns;
  KTreeData *ktd;
  KTreeResult *ktr;
  int i, timeMatch, timeCheck, timeData, t1, totalTime;
  SubstitutionMatrix *sm = NULL;

  totalTime = clock();
  
  version = VERSION;
  setprogname2("kerror");
  setAlignmentMode(KERROR);

  args = getArgs(argc, argv);
  query = subject = NULL;
  fp = NULL;
  if(args.h == 1){
    printUsage(version);
    return 0; 
  }else if(args.v == 1){
    printSplash(version);
    return 0;
  }else if(args.e == 1){
    printUsage(version);
    return -1;
  }
  /* get score matrix */
  if(args.m == NULL){
    sm = constructDnaScoreMatrix(sm, args.A, args.I);
  }else{
    fp = efopen(args.m, "r");
    sm = readSubstitutionMatrix(sm, fp);
  }
  if(fp != NULL)
    fclose(fp);
  timeData = timeCheck = timeMatch = 0;
  fpq = efopen(args.i,"r");
  fps = efopen(args.j,"r");
  /* get sequence data by looping over entries in the query and subject file */
  while(1){
    t1 = clock();
    query = getNextSequence2(fpq,query);
    if(query == NULL)
      break;
    /* ignore case */
    query->seq = strtoupper(query->seq);
    patterns = divideSequence(query,args.k+1);
    if(args.L)
	for(i=0;i<args.k+1;i++)
	    printf("#Fragment %d: %s\n",i+1,patterns[i].seq);
    root = ktree_constructRoot();
    for(i=0;i<args.k+1;i++){
      ktd = (KTreeData *)emalloc(sizeof(KTreeData));
      ktd->pos = patterns[i].qi;
      ktd->id = ktd->pos;
      ktree_addString(patterns[i].seq, ktd, root);
    }
    ktree_preprocessTree(root);
    rewind(fps);
    while(1){
      subject = getNextSequence2(fps,subject);
      if(subject == NULL) 
	break;
      /* ignore case */
      subject->seq = strtoupper(subject->seq);
      timeData += clock() - t1;
      t1 = clock();
      ktr = ktree_search(subject->seq, root);
      timeMatch += clock() - t1;
      t1 = clock();
      checking(ktr,query,subject,sm,"Plus / Plus",args);
      timeCheck += clock() - t1;
      if(args.r){
	t1 = clock();
	rev = revcomp(subject);
	timeData += clock() - t1;
	t1 = clock();
	ktr = ktree_search(rev->seq, root);
	timeMatch += clock() - t1;
	t1 = clock();
	checking(ktr,query,rev,sm,"Plus / Minus",args);
	timeCheck += clock() - t1;
      }
    }
  }

  printf("# Total time:        %.2fs\n",(float)(clock()-totalTime)/(float)CLOCKS_PER_SEC);
  printf("# Data manipulation: %.2fs\n",(float)timeData/(float)CLOCKS_PER_SEC);
  printf("# Matching:          %.2fs\n",(float)timeMatch/(float)CLOCKS_PER_SEC);
  printf("# Checking:          %.2fs\n",(float)timeCheck/(float)CLOCKS_PER_SEC); 

  if(query != NULL)
    freeSequence(query);
  if(subject != NULL)
    freeSequence(subject);
  
  fclose(fpq);
  fclose(fps);

  return 0;
}

void checking(KTreeResult *ktr, Sequence *query, Sequence *subject, SubstitutionMatrix *sm, char *strand, Args args){
  int lb, rb, prev_rb;
  KTreeResult *lp;
  Alignment *al;
  FILE *fp;

  if(args.o == NULL)
    fp = stdout;
  else
    fp = efopen(args.o, "w");

  prev_rb = 0;
  for(lp = ktr; lp != NULL; lp = lp->next){
    lb = lp->pos - (lp->data->pos + args.k + 1);
    if(lb < 0)
      lb = 0;
    rb = lp->pos + strlen(query->seq) - lp->data->pos + args.k - 1;
    if(rb >= strlen(subject->seq))
      rb = strlen(subject->seq) - 1;
    if(lb > prev_rb){
      prev_rb = rb;
      al = alignKerror(subject->seq,query->seq,lb,rb,sm,args);
      if(al != NULL){
	al->strand = strand;
	al->name1 = strdup2(query->id);
	al->name2 = strdup2(subject->id);
	al->length1 = strlen(query->seq);
	al->length2 = strlen(subject->seq);
	printAlignment(al,fp,sm,args);
	freeAlignment(al);
      }
    }
  }
  if(args.o != NULL)
    fclose(fp);
}
