#!/usr/bin/awk -f
# coalescent.awk
# Purpose: Coalescent simulation
# Reference: Hudson, R. R. (1990). Gene genealogies and the coalescent process. In: 
#   Futuyama, D. and Antonovics, J. (eds). Oxford Surveys in Evolutionary Biology, Vol. 7.
#   This classic paper contains neat C-code for coalescent simulations in its Appendix p. 41ff, 
#   which I have used with slight modifications.
# Date: January 17, 2013 (version 0.1)
#       March 3, 2017 (version 0.2)
BEGIN{
    program = "coalescent.awk"
    version = "0.2"
    if(h){
	printUsage(program)
    }
    if(v){
	printVersion(program, version)
    }
    srand(seed)
    for(ii=0;ii<numSamples;ii++){
	constructTree(sampleSize)
	if(printTree){
	    printNewickTree(2*sampleSize-1)
	    printf ";\n"
	}
	print "S= " segSites(sampleSize)
	if(printArrays)
	    printNodeArrays(sampleSize)
    }
}

function constructTree(n){
    # initialize node labels 1, 2, ..., 2*n-1
    for(i=1;i<2*n;i++)
	tree[i] = i
    # set times of leaves to zero
    for(i=1;i<=n;i++)
	time[i] = 0
    # set times of internal nodes
    t = 0
    for(i=n;i>1;i--){
	x = -2.*log(1.-rand())/i/(i-1)
	t += x
	time[2*n-i+1] = t
    }
    # construct tree topology
    for(i=n;i>1;i--){
	# get parent
	p = tree[2*n-i+1]
	# pick left child
	pick = int(i*rand())+1
	left[p] = tree[pick]
	# store parent
	parent[tree[pick]] = p
	# replace pick
	tree[pick] = tree[i]
	# pick right child
	pick = int((i-1)*rand())+1
	right[p] = tree[pick]
	# store parent
	parent[tree[pick]] = p
	# replace pick
	tree[pick] = p
    }
}

function segSites(n){
    generateMutations(n)
    s = 0
    for(i=1;i<2*n;i++)
	s += nMut[i]
    return s
}

function generateMutations(n){
    for(i=1;i<2*n;i++){
	if(parent[i]){
	    t = time[parent[i]] - time[i]
	    nMut[i] = poisson(t*theta/2)
	}
    }
}

# Attributed to Knuth, "Seminumerical Algorithms" (look up exact reference)
function poisson(lambda){

    l = exp(-lambda)
    k = 0
    p = 1
    while(p>l){
	k++
	p *= rand()
    }
    return k-1
}

######## Tree Printing ########
# Print node names followed by children, followed by parent
function printSimpleTree(i){
    if(i){
	printSimpleTree(left[i])
	if(left[i]){
	    printf i "; l: " left[i] "; r: " right[i] 
	}else
	    printf i
	if(parent[i])
	    printf "; p: " parent[i]
	printf "\n"
	printSimpleTree(right[i])
    }
}

# Print tree in nexus format
function printNewickTree(i){
    if(i){
	if(left[i]){  # internal node
	    printf("(")
	    printNewickTree(left[i])
	    printf(",")
	    printNewickTree(right[i])
	    printf(")")
	    printNode(i)
	}else{            # leaf
	    printNode(i)
	}
    }
}

function printNode(i){
    if(!left[i])
	printf("T%d",i)
    if(parent[i])
	printf(":%.5f",time[parent[i]]-time[i])
}

function printNodeArrays(n){
    printf("left\n")
    printTopArr(left,n)
    printf("right\n")
    printTopArr(right,n)
    printf("parent\n")
    printTopArr(parent,n)
    printf("time\n")
    printValArr(time,n,1)
    printf("nMut\n")
    printValArr(nMut,n,0)    
}

function printIndex(n){
    printf("1")
    for(i=2;i<2*n;i++)
	printf("\t%d",i)
    printf("\n")
}

function printTopArr(arr,n){
    printIndex(n)
    printf("%d",arr[1])
    for(i=2;i<2*n;i++)
	printf("\t%d",arr[i])
    printf("\n")
}

function printValArr(arr,n,f){
    printIndex(n)
    if(f)
	printf("%.2f",arr[1])
    else
	printf("%d",arr[1])
    for(i=2;i<2*n;i++)
	if(f)
	    printf("\t%.2f",arr[i])
	else
	    printf("\t%d",arr[i])
    printf("\n")
}

function printUsage(prog){
    	printf "Usage: awk -f %s -v seed=$RANDOM -v theta=<theta> -v sampleSize=<sampleSize> -v numSamples=<numSamples> [-v printTree=0|1 -v printArray=0|1 -v h=1 -v v=1]\n", prog
	printf "Purpose: Coalescent simulation\n"
	printf "Example: awk -f %s -v seed=$RANDOM -v theta=100 -v sampleSize=20 -v numSamples=2 -v printTree=1\n", prog
	printf "Options:\n"
	printf "\tseed\t\tseed for random number generator\n"
	printf "\ttheta\t\tpopulation mutation parameter\n"
	printf "\tsampleSize\tnumber of lineages at time zero\n"
	printf "\tnumSamples\tnumber of samples generated\n"
	printf "\t[h\t\thelp]\n"
	printf "\t[v\t\tversion]\n"
	exit
}

function printVersion(prog, ver){
    	printf "%s %s\n", prog, ver
	printf "Written by Bernhard Haubold.\n"
	printf "Distributed under the GNU General Public License.\n"
	printf "Please send bug reports to haubold@evolbio.mpg.de\n"
	exit
}
