/***** encode.c ***********************************
 * Description: Encoding step of Burrows-Wheeler
 *    Transform.
 * Reference: Donald Adjeroh, Timorhy Bell, and
 *    Amar Mukherjee (2008). The Burrows-Wheeler 
 *    Transform; Data Compression, Suffix Arrays, 
 *    and Pattern Matching. Springer.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * Date: Mon Mar  5 11:58:31 2012
 **************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "interface.h"
#include "eprintf.h"
#include "sequenceData.h"

char *globalText;
int globalTextLen;

int ringStrCmp(const void *v1, const void *v2);
void printRotations(char *text, int *indexArr, int n);
void printTransform(Args *args, char *text, char *header, int *indexArr, int n);
void printContext(Args *args, char *text, char *header, int *indexArr, int n);
void printRuns(Args *args, char *text, char *header, int *indexArr, int n);

void encode(Args *args, Sequence *seq){
  int *indexArr;
  int i;

  indexArr = (int *)emalloc(seq->len*sizeof(int));
  for(i=0;i<seq->len;i++)
    indexArr[i] = i;
  globalTextLen = seq->len;
  globalText = seq->seq;
  if(args->D){
    printf("Text:\n");
    printf("%s\n\n",seq->seq);
    printf("Rotations:\n");
    printRotations(globalText, indexArr, globalTextLen);
  }
  qsort(indexArr,globalTextLen,sizeof(int),ringStrCmp);
  if(args->D){
    printf("\nSorted Rotations:\n");
    printRotations(globalText, indexArr, globalTextLen);
    printf("\nTransform:\n");
    printTransform(args, globalText, NULL, indexArr, globalTextLen);
  }
  if(!args->D){
    if(!args->c){
      if(args->r)
	printRuns(args, globalText, seq->headers[0], indexArr, globalTextLen);
      else
	printTransform(args, globalText, seq->headers[0], indexArr, globalTextLen);
    }else
      printContext(args, globalText, seq->headers[0], indexArr, globalTextLen);
  }

  free(indexArr);
}

void printContext(Args *args, char *text, char *header, int *indexArr, int n){
  int i, j, start, end;
  
  if(args->p){
    start = args->p[0] > -1 ? args->p[0] : 0;
    end = args->p[1] + 2 < n ? args->p[1] + 2 : n;
  }else{
    start = 0;
    end = n+1;
  }

  for(i=start;i<end-1;i++){
    printf("%d: ",i+1-start);
    for(j=indexArr[i];j<indexArr[i]+args->l;j++)
      printf("%c",text[j%n]);
    printf("\n");
  }
  
}

void printRuns(Args *args, char *text, char *header, int *indexArr, int n){
  int i, j, o, start, end;
  int rowLen = 70;
  char c1, c2;
  char *count;

  count = (char *)emalloc(sizeof(char)*256);
  for(i=0;i<256;i++)
    count[i] = 0;

  if(header)
    printf("%s - bwt\n",header);
  o = n-1;
  j = 0;
  if(args->p){
    start = args->p[0] > -1 ? args->p[0] : 0;
    end = args->p[1] + 2 < n ? args->p[1] + 2 : n;
  }else{
    start = 0;
    end = n+1;
  }
  c1 = text[(indexArr[start]+o)%n];
  count[(int)c1] = 1;
  for(i=start+1;i<end-1;i++){
    c2 = text[(indexArr[i]+o)%n];
    if(c2 != c1){
      if(count[(int)c1] > 1)
	printf("%c%d",c1,count[(int)c1]);
      else
	printf("%c",c1);
      j += (log(count[(int)c1])/log(10)+1) + 1;
      count[(int)c1] = 0;
      if(j>=rowLen){
	j = 0;
	printf("\n");
      }
    }
    count[(int)c2]++;
    c1 = c2;
  }
  if(count[(int)c1] > 1)
    printf("%c%d",c1,count[(int)c1]);
  else
    printf("%c",c1);
  printf("\n");
  free(count);
}


void printTransform(Args *args, char *text, char *header, int *indexArr, int n){
  int i, j, o, start, end;
  int rowLen = 70;

  if(header)
    printf("%s - bwt\n",header);
  o = n-1;
  j = 0;
  if(args->p){
    start = args->p[0] > -1 ? args->p[0] : 0;
    end = args->p[1] + 2 < n ? args->p[1] + 2 : n;
  }else{
    start = 0;
    end = n+1;
  }
    
  for(i=start;i<end-1;i++){
    j++;
    printf("%c",text[(indexArr[i]+o)%n]);
    if(j==rowLen){
      j = 0;
      printf("\n");
    }
  }
  if(j)
    printf("\n");
}

void printRotations(char *text, int *indexArr, int n){
  int i, j;

  for(i=0;i<n;i++){
    for(j=0;j<n;j++)
      printf("%c",text[(indexArr[i]+j)%n]);
    printf("\n");
  }

}

int ringStrCmp(const void *v1, const void *v2){
  int a, b, i;

  a = *(int *)v1;
  b = *(int *)v2;
  i = 0;

  while(globalText[a % globalTextLen] == globalText[b % globalTextLen] && i < globalTextLen){
    a++;
    b++;
    i++;
  }

  if(globalText[a] == globalText[b])
    return 0;
  else if(globalText[a] < globalText[b])
    return -1;
  else
    return 1;
}


