/***** construct.c ************************************************
 * Description: Routines for constructing a keyword tree.
 * Reference: Gusfield, D. (1997). Algorithms on Strings, Trees,
 *    and sequences; Computer Science and Computational Biology.
 *    CUP, p. 52ff.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Wed Sep  8 10:20:28 2004.
 *****************************************************************/
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "eprintf.h"
#include "keywordMatcher.h"

int nodeId = 0;

/*
 * Purpose: generate node.
 * Parameter: none
 * Returns: pointer to allocated node
 */
KNode *getNode(){
  KNode *np;

  np = (KNode *)emalloc(sizeof(KNode));
  np->id = -1;
  np->nId = nodeId++;
  np->depth = 0;
  np->child = NULL;
  np->sibling = NULL;
  np->fail = NULL;
  np->out = NULL;
  np->c = '\0';

  return np;
}


/*
 * Generate keyword tree from patterns in Sequence format. Otherwise identical
 * to plain createKeywordTree.
 */
KNode *createSequenceKeywordTree(Sequence *patterns, int numPat, int numChar){
  KNode *root;
  int i;

  root = getNode();
  for(i=0; i<numPat; i++)
    addString(patterns[i].seq, i, root);
  preprocessTree(root);
  return root;
}

/*
 * Generate a preprocessed keyword tree ready for matching.
 * char **patterns: set of patterns to be included in the tree
 * int numPat: number of patterns
 * int numChar: total number of characters contained in the pattern set
 */
KNode *createKeywordTree(char **patterns, int numPat, int numChar){
  KNode *root;
  int i; 
  
  root = getNode();
  for(i=0; i<numPat; i++)
    addString(patterns[i], i, root);
  preprocessTree(root);

  return root;
}

/*
 * Add a string to the keyword tree. Because we are dealing with 
 * a trie we cannot treat every node as branching.
 */
void addString(char *str, int id, KNode *root){
  KNode *np, *child, *parent;
  int stringLength;
  int depth;
  depth = 0;
  np = root; 
  
  stringLength= strlen(str);
  parent = np;
  while((np = findChild(np,*str)) != NULL && depth < stringLength){
    str++;
    depth++;
    parent = np;
  }
  if(depth == stringLength && np != NULL){
    /* make sure pattern list is non-redundant */
    if(np->id == -1){
      np->id = id;
    }else{
      printf("ERROR [keywordTree]: Attempted to add two identical strings to keyword tree.\n");
      printf("                     The program only accepts non-redundant pattern lists.\n");
      exit(0);
    }
  }else{
    np = parent;
    while(depth < stringLength){
      child = getNode();
      child->c = *str++;
      child->depth = ++depth;
      addChild(parent,child);
      parent = child;
    }
    parent->id = id;
  }
}

void addChild(KNode *parent, KNode *child){
  KNode *np;

  if(parent->child == NULL){
    parent->child = child;
  }else{
    np = parent->child;
    while(np->sibling != NULL){
      np = np->sibling;
    }
    np->sibling = child;
  }
}

/*
 * Purpose: Find a child node of parent whose edge is labelled with c.
 */
KNode *findChild(KNode *parent, char c){
  KNode *np;

  if(parent->child == NULL){
    return NULL;
  }
  np = parent->child;
  while(np->c != c){
    if(np->sibling == NULL){
      break;
    }
      np = np->sibling;
  }
  if(np->c == c){
    return np;
  }else{
    return NULL;
  }
}
