/***** streeComputations.c ****************************************
 * Description: Computations on the suffix tree
 * Author: Bernhard Haubold, bernhard.haubold@fh-weihenstephan.de
 * File created on Sun May 15 18:58:08 2005.
 *****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "stree_strmat.h"
#include "stree_ukkonen.h"
#include "eprintf.h"
#include "interface.h"
#include "sequenceData.h"
#include "stringUtil.h"
#include "repeater.h"

Args *args;

void findRepeats(SUFFIX_TREE tree){
  int i;

  /* initialize character map */
  for(i=0;i<256;i++)
      tree->charMap[i] = 0;
  stree_traverse(tree, initializeStringDepth, NULL);
  stree_traverse(tree, stringDepth, NULL);
  tree->maxRepeatLen = INT_MIN;
  tree->minRepeatLen = INT_MAX;
  if(args->a || (args->m == INT_MAX && args->M == INT_MIN)) /* look for maximal or longest repeats? */
    stree_traverse(tree, leftDivergence, NULL);
  else
    stree_traverse(tree, setLeftDivergent, NULL);
  if(tree->minRepeatLen == 0)
    tree->minRepeatLen = tree->maxRepeatLen;
  if(args->q==2)
    fprintf(tree->fpout,"|%d<=l<=%d\n", tree->minRepeatLen, tree->maxRepeatLen);
  else{
    tree->isFirstOutput = 1;
    if(args->q<2)
      fprintf(tree->fpout,"#len|strId:pos_1|...|strId:pos_n");
    if(args->q<1)
      fprintf(tree->fpout,"|seq");
    fprintf(tree->fpout,"\n");
    stree_traverse(tree, output, NULL);
  }
}


int output(SUFFIX_TREE tree, STREE_NODE node){
  int i, endPos;
  
  if(!node->isaleaf && node->isLeftDivergent){
    tree->substrLen = node->stringDepth;
    if((node->stringDepth >= args->m && node->stringDepth <= args->M) \
       || (node->stringDepth >= args->m && args->M == INT_MIN)        \
       || (args->m == INT_MAX && node->stringDepth <= args->M)){
      if(args->q <= 1){
	fprintf(tree->fpout, "%d", node->stringDepth);
	tree->isFirstOutput = 0;
      }
      tree->isFirstPrint = 1;
      stree_traverse_subtree(tree, node, printPositions, NULL);
      if(args->q == 0){
	fprintf(tree->fpout, "|");
	endPos = tree->startPos + tree->substrLen - 1;
	for(i=tree->startPos; i<=endPos; i++)
	  fprintf(tree->fpout,"%c",tree->strings[tree->strid][i]);
      }
      fprintf(tree->fpout,"\n");
    }else if(node->stringDepth >= tree->maxRepeatLen && args->m == INT_MAX && args->M == INT_MIN){
      fprintf(tree->fpout, "%d", node->stringDepth);
      tree->isFirstPrint = 1;
      stree_traverse_subtree(tree, node, printPositions, NULL);
      if(args->q == 0){
	fprintf(tree->fpout, "|");
	endPos = tree->startPos + tree->substrLen - 1;
	for(i=tree->startPos; i<=endPos; i++)
	  fprintf(tree->fpout,"%c",tree->strings[tree->strid][i]);
      }
      fprintf(tree->fpout,"\n");
    }
  }
  return 0;
}

int printPositions(SUFFIX_TREE tree, STREE_NODE node){
  STREE_NODE parent;
  STREE_LEAF leaf;
  int pos, id, leafnum;
  char **string = NULL;

  string = (char **)emalloc(tree->nextslot*sizeof(char *));
  if(!node->isaleaf){
    leafnum = 1;
    while(stree_get_leaf(tree, node, leafnum++, string, &pos, &id)){
      if(args->r && id % 2 == 1){
	pos = tree->lengths[id] - (pos + node->stringDepth);
      }else{
	pos = pos + 1;
      }
      if(tree->nextslot > 1){
	if(args->r && (id+1) % 2 == 0)        /* reverse strand */
	  fprintf(tree->fpout, "|r%d:%d", id+1, pos + 1);
	else                                  /* forward strand */
	  fprintf(tree->fpout, "|f%d:%d", id+1, pos);
	if(tree->isFirstPrint){
	  tree->isFirstPrint = 0;
	}
      }else{
	if(args->r && (id+1) % 2 == 0)          /* reverse strand */
	  fprintf(tree->fpout, "|%d", pos + 1);
	else{                                  /* forward strand */
	  fprintf(tree->fpout, "|%d", pos);
	}
	if(tree->isFirstPrint)
	  tree->isFirstPrint = 0;
      }
    }
    tree->endPos = pos - 1 + node->stringDepth -1;
  }else{
    parent = stree_get_parent(tree,node);
    leaf = (STREE_LEAF)node;
    id = leaf->strid;
    if(args->r && id % 2 == 1){
      pos = tree->lengths[id] - (leaf->pos + parent->stringDepth);
    }else{
      pos = leaf->pos + 1;
    }
    if(tree->nextslot > 1){
      if(args->r && (id+1) % 2 == 0) /* reverse strand */
	fprintf(tree->fpout, "|r%d:%d", id+1, pos + 1);
      else                                  /* forward strand */
	fprintf(tree->fpout, "|f%d:%d", id+1, pos);
      if(tree->isFirstPrint){
	tree->isFirstPrint = 0;
      }
    }else{
      if(args->r && (id+1) % 2 == 0) /* reverse strand */
	fprintf(tree->fpout, "|%d", pos + 1);
      else                                  /* forward strand */
	fprintf(tree->fpout, "|%d", pos);
      if(tree->isFirstPrint){
	tree->isFirstPrint = 0;
      }
    }
    tree->endPos = pos - 1 + parent->stringDepth -1;
  }
  tree->startPos = pos - 1;
  tree->strid = id;
  return 0;
}

/* setLeftDivergent: set all nodes left divergent 
 *
 */
int setLeftDivergent(SUFFIX_TREE tree, STREE_NODE node){

  if(!node->isaleaf)
    node->isLeftDivergent = 1;

  return 0;
}

/* leftDivergence: determine whether node is left divergent; 
 *   call preorder
 */
int leftDivergence(SUFFIX_TREE tree, STREE_NODE node){
  STREE_NODE child;
  STREE_NODE np;
  STREE_LEAF leaf;
  int leafCount, i;

  if(!node->isaleaf){
    child = stree_get_children(tree, node);
    leafCount = 0;
    for(np=child; np != NULL; np = np->next){
      if(np->isaleaf){
	leaf = (STREE_LEAF)np;
	if(tree->charMap[(int)tree->strings[leaf->strid][leaf->pos]] == 0){
	  tree->charMap[(int)tree->strings[leaf->strid][leaf->pos]] = 1;
	  tree->charIndex[leafCount++] = tree->strings[leaf->strid][leaf->pos];
	}else{
	  break;
	}
      }else if(!np->isLeftDivergent){
	break;
      }
      node->isLeftDivergent = 1;
      if(node->stringDepth > tree->maxRepeatLen)
	tree->maxRepeatLen = node->stringDepth;
      if(node->stringDepth < tree->minRepeatLen)
	tree->minRepeatLen = node->stringDepth;
      /* reset character map */
      for(i=0;i<leafCount;i++)
	tree->charMap[tree->charIndex[i]] = 0;
    }
  }
  return 0;
}

/* stringDepth: compute the number of characters on the path 
 *   from the root to an internal node
 *   call preorder
 */
int stringDepth(SUFFIX_TREE tree, STREE_NODE node){
  STREE_NODE parent;

  parent = stree_get_parent(tree,node);
  if(parent != NULL && !node->isaleaf){
    node->stringDepth = parent->stringDepth + stree_get_edgelen(tree,node);
  }
  return 0;
}

/* initializeStringDepth: set all string depths to zero
 */
int initializeStringDepth(SUFFIX_TREE tree, STREE_NODE node){

  if(!node->isaleaf){
    node->stringDepth = 0;
    node->isLeftDivergent = 0;
  }

  return 0;
}

/* setArgs: cludge to make the argument structure available to the suffix tree computations */
void setArgs(Args *arguments){
  args = arguments;
}


 
