/***** sequenceData.c *********************************************
 * Description: Collection of routines for reading and 
 * manipulating sequence data.
 * Author: Bernhard Haubold, bernhard.haubold@fh-weihenstephan.de
 * File created on Sun Jun  6 10:34:31 2004.
 *****************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "eprintf.h"
#include "sequenceData.h"
#include "stringUtil.h"

/* getDnaDictionary: create DNA dictionary */
int *getDnaDictionary(int *dic){
  int i;

  if(dic == NULL)
    dic = (int *)malloc((DICSIZE+1)*sizeof(int));

  for(i=0; i<DICSIZE; i++)
    dic[i] = 0;

  dic['a'] = 0;   /* a */
  dic['c'] = 1;   /* c */
  dic['g'] = 2;   /* g */
  dic['t'] = 3;   /* t */
  dic['A'] = 0;   /* A */
  dic['C'] = 1;   /* C */
  dic['G'] = 2;   /* G */
  dic['T'] = 3;   /* T */
  dic['r'] = dic['g'];
  dic['R'] = dic['g'];
  dic['y'] = dic['t'];
  dic['Y'] = dic['t'];
  dic['m'] = dic['a'];
  dic['M'] = dic['a'];
  dic['k'] = dic['g'];
  dic['K'] = dic['g'];
  dic['s'] = dic['g'];
  dic['S'] = dic['g'];
  dic['w'] = dic['a'];
  dic['W'] = dic['a'];
  dic['h'] = dic['a'];
  dic['H'] = dic['a'];
  dic['b'] = dic['g'];
  dic['B'] = dic['g'];
  dic['v'] = dic['g'];
  dic['V'] = dic['g'];
  dic['d'] = dic['g'];
  dic['D'] = dic['g'];
  dic['n'] = dic['g'];
  dic['N'] = dic['g'];
  dic['u'] = dic['t'];
  dic['U'] = dic['t'];

  return dic;
}


void rev(char *seq){
  char *newStr;
  int i, j, l;

  l = strlen(seq);

  newStr = (char *)malloc(sizeof(char)*l);
  j = 0;
  for(i=l-1;i>-1;i--)
    newStr[j++] = seq[i];
  for(i=0;i<l;i++)
    seq[i] = newStr[i];
  free(newStr);
}

void comp(char *seq){
  int i, l;
  char c;
  
  l = strlen(seq);
  c = '\0';
  for(i=0;i<l;i++){
    c = seq[i];
    switch(c){
    case 'a':
      seq[i] ='t';
      break;
    case 'c':
      seq[i] ='g';
      break;
    case 'g':
      seq[i] ='c';
      break;
    case 't':
      seq[i] ='a';
      break;
    case 'A':
      seq[i] ='T';
      break;
    case 'C':
      seq[i] ='G';
      break;
    case 'G':
      seq[i] ='C';
      break;
    case 'T':
      seq[i] ='A';
      break;
    default:
      seq[i] = c;
      break;
    }
  }
  seq[i] = '\0';
}

/* reverse and complement a sequence */
void revcomp(char *seq){
  rev(seq);
  comp(seq);
}

/* Get next sequence from an open data stream in FASTA format; this stream may be the stdin */
Sequence *getNextSequence(FILE *fp){
  static char *line = NULL;
  static int last = 0;
  Sequence *sequence;
  int seqlen, seqi, i, l;
  int currentBuffer;

  if(last)
    return NULL;

  if(line == NULL){
    line = (char *)emalloc((SEQLINE+2)*sizeof(char));
    line = fgets(line,SEQLINE,fp);
  }
  /* make a sequence object */
  sequence = (Sequence *) emalloc(sizeof(Sequence));
  /* allocate memory for sequence id */
  sequence->id = (char *)emalloc((strlen(line)+1)*sizeof(char));
  sequence->id[0] = '\0';
  /* copy sequence id */
  strcpy(sequence->id,line);
  /* allocate memory for sequence string */
  sequence->seq = (char *)emalloc((SEQBUFFER+1)*sizeof(char));
  sequence->seq[0] = '\0';
  seqlen = 0;
  currentBuffer = SEQBUFFER;
  seqi = 0;
  while((line=fgets(line,SEQLINE,fp)) != NULL){
    if(strstr(line,">") != NULL){
      sequence->seq = (char *) erealloc(sequence->seq,strlen(sequence->seq)+1);
      free(line);
      return sequence;
    }
    if(strlen(line)>SEQLINE){
      printf("error in getNextSequence: cannot deal with lines longer than %d bp.\n",SEQLINE);
      printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
      exit(2);
    }
    l = strlen(line);
   /* disregard the final carriage return */
    if(line[l-1] == '\n')
      l--;
    seqlen += l; 
    if(seqlen > currentBuffer){
      currentBuffer += SEQBUFFER;
      sequence->seq = (char *)erealloc(sequence->seq,currentBuffer);
    }
    for(i=0;i<l;i++)
      sequence->seq[seqi++] = line[i];
    sequence->seq[seqi] = '\0';
    /* sequence->seq = strncat(sequence->seq,line,strlen(line)-1); */
  }
  sequence->seq = (char *) erealloc(sequence->seq,strlen(sequence->seq)+1);
  last = 1;
  free(line);
  return sequence;
}

/* read FASTA file and store data; return size of data set */
Sequence *readFasta(char *file, Sequence *data, int *size, int *residues){
    FILE *fp;
    char *line;
    int first, l, numentry, maxseqlen, seqlen;
    int *lengthptr;
    Sequence *od;

    line = (char *)malloc((SEQLINE+2)*sizeof(char));

    /* open input file NB: this has got to be an external file, otherwise the rewind fails */
    fp = fopen(file, "r");
    if(fp == NULL){
	fprintf(stderr,"error in readFasta: could not open %s\n.",file);
	exit(1);
    }
    
    /* count number of entries and determine length of longest sequence */
    numentry = 0;
    maxseqlen = 0;
    seqlen = 0;
    *residues = 0;
    while((line=fgets(line,SEQLINE+2,fp)) != NULL){
	if(strlen(line)>SEQLINE){
	    printf("error in readFasta: cannot deal with lines longer than %d bp.\n",SEQLINE);
	    printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
	    exit(2);
	}
	if(strstr(line,">") != NULL){
	    numentry++;
	    if(seqlen > maxseqlen)
		maxseqlen = seqlen;
	    seqlen = 0;
	}else{
	    l = strlen(line) - 1;
	    seqlen += l; 
	    *residues += l;
	}
    }
    *size = numentry;
    /* allocate memory */
    data = (Sequence *)malloc(numentry*sizeof(Sequence));
    lengthptr = (int *)malloc(numentry*sizeof(int));
    line = (char *)malloc((SEQLINE+2)*sizeof(char));
    od = data;
    rewind(fp);
    /* Determine the length of each input sequence */
    lengthptr--;
    while((line=fgets(line,SEQLINE+2,fp)) != NULL){
	if(strstr(line,">") != NULL){
	    lengthptr++;
	    *lengthptr=0;
	}else{
	    l = strlen(line) - 1;
	    *lengthptr += l;
	}
    }

    line = (char *)malloc((SEQLINE+2)*sizeof(char));
    rewind(fp);

    first = 1;
    lengthptr -= numentry;
    data--;
    while((line=fgets(line,SEQLINE,fp)) != NULL){
	line = chomp(line);
	if(strstr(line,">") != NULL){           /* found new entry */
	    if(first == 0){
		data->seq -= *lengthptr;        /* readjust sequence pointer position */
	    }else{
		first = 0;
	    }
	    lengthptr++;
	    data++;
	    data->seq = (char *)malloc(*lengthptr * sizeof(char));
	    data->id = (char *)malloc((strlen(line)+1)*sizeof(char));
	    strcpy(data->id,line);
	}else{
	    strcpy(data->seq,line);
	    data->seq += strlen(line);
	}
    }
    /* readjust last sequence pointer */
    data->seq -= *lengthptr;
    free(line);
    return od;
}

/* freeSequence: free the data structure Sequence */
Sequence *freeSequence(Sequence *seq){
  free(seq->id);
  free(seq->seq);
  free(seq);
  return seq;
}
