/***** sequenceData.c *********************************************
 * Description: Collection of routines for reading and 
 * manipulating sequence data.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Sun Jun  6 10:34:31 2004.
 *****************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "eprintf.h"
#include "sequenceData.h"
#include "stringUtil.h"

static int last = 0;

/* everything that is not [acgtACGT] is flagged by a -1 */
int *getRestrictedDnaDictionary(int *dic){
  int i;

  if(dic == NULL)
    dic = (int *)malloc((DICSIZE+1)*sizeof(int));

  for(i=0; i<DICSIZE; i++)
    dic[i] = -1;

  dic['a'] = 0;   /* a */
  dic['c'] = 1;   /* c */
  dic['g'] = 2;   /* g */
  dic['t'] = 3;   /* t */
  dic['A'] = 0;   /* A */
  dic['C'] = 1;   /* C */
  dic['G'] = 2;   /* G */
  dic['T'] = 3;   /* T */

  return dic;

}

/* getDnaDictionary: create DNA dictionary */
int *getDnaDictionary(int *dic){
  int i;

  if(dic == NULL)
    dic = (int *)malloc((DICSIZE+1)*sizeof(int));

  for(i=0; i<DICSIZE; i++)
    dic[i] = 0;

  dic['a'] = 0;   /* a */
  dic['c'] = 1;   /* c */
  dic['g'] = 2;   /* g */
  dic['t'] = 3;   /* t */
  dic['A'] = 0;   /* A */
  dic['C'] = 1;   /* C */
  dic['G'] = 2;   /* G */
  dic['T'] = 3;   /* T */
  dic['r'] = dic['g'];
  dic['R'] = dic['g'];
  dic['y'] = dic['t'];
  dic['Y'] = dic['t'];
  dic['m'] = dic['a'];
  dic['M'] = dic['a'];
  dic['k'] = dic['g'];
  dic['K'] = dic['g'];
  dic['s'] = dic['g'];
  dic['S'] = dic['g'];
  dic['w'] = dic['a'];
  dic['W'] = dic['a'];
  dic['h'] = dic['a'];
  dic['H'] = dic['a'];
  dic['b'] = dic['g'];
  dic['B'] = dic['g'];
  dic['v'] = dic['g'];
  dic['V'] = dic['g'];
  dic['d'] = dic['g'];
  dic['D'] = dic['g'];
  dic['n'] = dic['g'];
  dic['N'] = dic['g'];
  dic['u'] = dic['t'];
  dic['U'] = dic['t'];

  return dic;
}


/* reverse and complement a sequence; return pointer to reverse-complemented sequence */
Sequence *revcomp(Sequence *seq){
  int i,j,n;
  char c;
  Sequence *newSeq;

  if(seq == NULL)
    return NULL;

  newSeq = (Sequence *)emalloc(sizeof(Sequence));

  n = strlen(seq->seq);
  newSeq->seq = (char *) emalloc((n+1)*sizeof(char));

  newSeq->id = strdup2(seq->id);  
  j=0;
  for(i = n-1; i >= 0; i--){
    c = seq->seq[i];
    switch(c){
    case 'a':
      newSeq->seq[j++] ='t';
      break;
    case 'c':
      newSeq->seq[j++] ='g';
      break;
    case 'g':
      newSeq->seq[j++] ='c';
      break;
    case 't':
      newSeq->seq[j++] ='a';
      break;
    case 'A':
      newSeq->seq[j++] ='T';
      break;
    case 'C':
      newSeq->seq[j++] ='G';
      break;
    case 'G':
      newSeq->seq[j++] ='C';
      break;
    case 'T':
      newSeq->seq[j++] ='A';
      break;
    default:
      newSeq->seq[j++] =c;
      break;
    }
  }
  newSeq->seq[n]='\0';
  seq = newSeq;
  return newSeq; 
}

void resetSequenceReader(){
  last = 0;
}

/* Get next sequence from an open data stream in FASTA format; this stream may be the stdin */
Sequence *getNextSequence2(FILE *fp, Sequence *sequence){
  int seqlen, seqi, i, l;
  int currentBuffer;
  Sequence *sequence2;

  if(sequence == NULL){
    sequence = (Sequence *)emalloc(sizeof(Sequence));
    sequence->line = (char *)emalloc((SEQLINE+2)*sizeof(char));
    sequence->line = chomp(fgets(sequence->line,SEQLINE,fp));
    /* allocate memory for sequence id */
    sequence->id = (char *)emalloc((strlen(sequence->line)+1)*sizeof(char));  
    /* copy sequence id */
    strcpy(sequence->id,sequence->line);
  }else if(sequence->isLast){
    freeSequence(sequence);
    return NULL;
  }else{
    sequence2 = (Sequence *)emalloc(sizeof(Sequence));
    sequence2->line = (char *)emalloc((SEQLINE+2)*sizeof(char));
    /* allocate memory for sequence id */
    sequence2->id = (char *)emalloc((strlen(sequence->line)+1)*sizeof(char));  
    /* copy sequence id */
    strcpy(sequence2->id,sequence->line);
    freeSequence(sequence);
    sequence = sequence2;
  }    
  sequence->isLast = 0;
  /* allocate memory for sequence string */
  sequence->seq = (char *)emalloc((SEQBUFFER+1)*sizeof(char));
  seqlen = 0;
  currentBuffer = SEQBUFFER;
  seqi = 0;
  while((sequence->line=fgets(sequence->line,SEQLINE,fp)) != NULL){
    sequence->line = chomp(sequence->line);
    if(strstr(sequence->line,">") != NULL){
      sequence->seq = (char *) erealloc(sequence->seq,strlen(sequence->seq)+1);
      return sequence;
    }
    if(strlen(sequence->line)>SEQLINE){
      printf("error in getNextSequence: cannot deal with lines longer than %d bp.\n",SEQLINE);
      printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
      exit(2);
    }
    l = strlen(sequence->line);
    seqlen += l; 
    if(seqlen > currentBuffer){
      currentBuffer += SEQBUFFER;
      if((sequence->seq = (char *)realloc(sequence->seq,currentBuffer)) == NULL){
	printf("ERROR [wordCounter]: memory allocation failed\n");
	    assert(0);
      }
    }
    for(i=0;i<l;i++){
      sequence->seq[seqi++] = sequence->line[i];
    }
  }
  sequence->seq[seqi] = '\0';
  sequence->seq = (char *) realloc(sequence->seq,strlen(sequence->seq)+1);
  sequence->isLast = 1;
  return sequence;
}


/* Get next sequence from an open data stream in FASTA format; this stream may be the stdin */
Sequence *getNextSequence(FILE *fp){
  static char *line = NULL;
  Sequence *sequence;
  int seqlen, seqi, i, l;
  int currentBuffer;

  if(last){
    return NULL;
  }

  if(line == NULL){
    line = (char *)malloc((SEQLINE+2)*sizeof(char));
    line = chomp(fgets(line,SEQLINE,fp));
  }

  /* make a sequence object */
  sequence = (Sequence *) malloc(sizeof(Sequence));
  /* allocate memory for sequence id */
  sequence->id = (char *)malloc((strlen(line)+1)*sizeof(char));
  /* copy sequence id */
  strcpy(sequence->id,line);
  /* allocate memory for sequence string */
  sequence->seq = (char *)malloc((SEQBUFFER+1)*sizeof(char));

  seqlen = 0;
  currentBuffer = SEQBUFFER;
  seqi = 0;
  while((line=fgets(line,SEQLINE,fp)) != NULL){
    line = chomp(line);
    if(strstr(line,">") != NULL){
      sequence->seq = (char *) realloc(sequence->seq,strlen(sequence->seq)+1);
      return sequence;
    }
    if(strlen(line)>SEQLINE){
      printf("error in getNextSequence: cannot deal with lines longer than %d bp.\n",SEQLINE);
      printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
      exit(2);
    }
    l = strlen(line);
    seqlen += l; 
    if(seqlen > currentBuffer){
      currentBuffer += SEQBUFFER;
      if((sequence->seq = (char *)realloc(sequence->seq,currentBuffer)) == NULL){
	printf("ERROR [wordCounter]: memory allocation failed\n");
	    assert(0);
      }
    }
    for(i=0;i<l;i++){
      sequence->seq[seqi++] = line[i];
    }
  }
  sequence->seq = (char *) realloc(sequence->seq,strlen(sequence->seq)+1);
  last = 1;
  return sequence;
}

/* read FASTA file and store data; return size of data set */
Sequence *readFasta(char *file, Sequence *data, int *size, int *residues){
    FILE *fp;
    char *line;
    int first, l, numentry, maxseqlen, seqlen;
    int *lengthptr;
    Sequence *od;

    line = (char *)malloc((SEQLINE+2)*sizeof(char));

    /* open input file NB: this has got to be an external file, otherwise the rewind fails */
    fp = fopen(file, "r");
    if(fp == NULL){
	fprintf(stderr,"error in readFasta: could not open %s\n.",file);
	exit(1);
    }
    
    /* count number of entries and determine length of longest sequence */
    numentry = 0;
    maxseqlen = 0;
    seqlen = 0;
    *residues = 0;
    while((line=fgets(line,SEQLINE+2,fp)) != NULL){
	if(strlen(line)>SEQLINE){
	    printf("error in readFasta: cannot deal with lines longer than %d bp.\n",SEQLINE);
	    printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
	    exit(2);
	}
	if(strstr(line,">") != NULL){
	    numentry++;
	    if(seqlen > maxseqlen)
		maxseqlen = seqlen;
	    seqlen = 0;
	}else{
	    l = strlen(line) - 1;
	    seqlen += l; 
	    *residues += l;
	}
    }
    *size = numentry;
    /* allocate memory */
    data = (Sequence *)emalloc(numentry*sizeof(Sequence));
    lengthptr = (int *)emalloc(numentry*sizeof(int));
    line = (char *)emalloc((SEQLINE+2)*sizeof(char));
    od = data;
    rewind(fp);
    /* Determine the length of each input sequence */
    lengthptr--;
    while((line=fgets(line,SEQLINE+2,fp)) != NULL){
	if(strstr(line,">") != NULL){
	    lengthptr++;
	    *lengthptr=0;
	}else{
	    l = strlen(line);
	    *lengthptr += l;
	}
    }

    line = (char *)emalloc((SEQLINE+2)*sizeof(char));
    rewind(fp);

    first = 1;
    lengthptr -= numentry;
    data--;
    while((line=fgets(line,SEQLINE,fp)) != NULL){
      line = chomp(line);
	printf("line: %s\n",line);
	if(strstr(line,">") != NULL){           /* found new entry */
	    if(first == 0){
		data->seq -= *lengthptr;        /* readjust sequence pointer position */
	    }else{
		first = 0;
	    }
	    lengthptr++;
	    data++;
	    data->seq = (char *)emalloc(*lengthptr * sizeof(char));
	    data->id = (char *)emalloc((strlen(line)+1)*sizeof(char));
	    strcpy(data->id,line);
	}else if (strlen(line) > 0){
	    strcpy(data->seq,line);
	    data->seq += strlen(line);
	}
    }
    /* readjust last sequence pointer */
    data->seq -= *lengthptr;
    free(line);
    return od;
}

/* freeSequence: free the data structure Sequence */
Sequence *freeSequence(Sequence *seq){
  free(seq->id);
  free(seq->seq);
  free(seq);
  return seq;
}

/* divideSequence: divide sequence into k segments */
Sequence *divideSequence(Sequence *seq, int k){
  int r, i, j, m, qi;
  Sequence *sp;

  r = strlen(seq->seq)/k;
  sp = (Sequence *)emalloc(k*sizeof(Sequence));
  qi = 0;
  
  /* treat the first k-1 segments */
  for(i=0;i<k-1;i++){
    sp[i].seq = (char *)emalloc((r+1)*sizeof(char));
    sp[i].id = (char *)emalloc(10*sizeof(char));
    itoa(i,sp[i].id);
    sp[i].qi = qi;
    for(j=0;j<r;j++){
      sp[i].seq[j]=seq->seq[qi++];
    }
    sp[i].seq[j] = '\0';
  }
  /* treat the last segment */
  sp[i].seq = (char *)emalloc((strlen(seq->seq)-qi+2)*sizeof(char));
  sp[i].qi = qi;
  sp[i].id = (char *)emalloc(10*sizeof(char));
  itoa(i,sp[i].id);
  m = 0;
  for(j=qi;j<strlen(seq->seq);j++)
    sp[i].seq[m++] = seq->seq[j];
  sp[i].seq[m] = '\0';
  return sp;
}

/* getSubstrings: get all substrings of length w */
Sequence *getSubstrings(Sequence *seq, int w){
  int i, j, seqLen, seqLen2;
  Sequence *sp;

  seqLen = strlen(seq->seq);
  if(seqLen-w+1 < 0){
    fprintf(stderr,"ERROR [getSubstrings]: sequence length - word length + 1 = %d < 0\n",seqLen-w+1);
    exit(-1);
  }
  sp = (Sequence *)emalloc((seqLen-w+1)*sizeof(Sequence));
  /* generate patterns of length w */
  seqLen2 = seqLen - w + 1;
  for(i=0;i<seqLen2;i++){
    sp[i].seq = (char *)emalloc((w+1)*sizeof(char));
    sp[i].id = (char *)emalloc(10*sizeof(char));
    itoa(i,sp[i].id);
    sp[i].qi = i;
    for(j=0;j<w;j++){
      sp[i].seq[j]=seq->seq[i+j];
    }
    sp[i].seq[j] = '\0';
  }
  return sp;
}
