/***** interface.c ************************************************
 * Description: Routine for gathering arguments from the command
 *              line.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Sun Jun 20 13:12:10 2004.
 *****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "eprintf.h"
#include "interface.h"

static struct option long_options[] = {
  {"coverage", required_argument, 0, 'c'},
  {"mean-fragment-length", required_argument, 0, 'l'},
  {"mean-read-length", required_argument, 0, 'R'},
  {"sd-fragment-length", required_argument, 0, 'd'},
  {"sd-read-length", required_argument, 0, 'D'},
  {"minimum-coverage", required_argument, 0, 'm'},
  {"sequencing-error-rate", required_argument, 0, 'E'},
  {"output-line-length", required_argument, 0, 'L'},
  {"random-seed", required_argument, 0, 's'},
  {"start-position", required_argument, 0, 'S'},
  {"paired-end-reads", no_argument, 0, 'P'},
  {"print-profiles", no_argument, 0, 'p'},
  {"single-contig", no_argument, 0, 'C'},
  {"circular-genome", no_argument, 0, 'r'},
  {"alignment", no_argument, 0, 'H'},
  {"help", no_argument, 0, 'h'},
  {"version", no_argument, 0, 'v'}
};


Args *
get_args (int argc, char *argv[])
{
  Args *args;
  int c;
  int opti;
  char *opt_string;

  opt_string = "c:l:d:m:E:L:s:S:R:D:PCprahv";

  args = (Args *) emalloc (sizeof (Args));
  args->c = DEFAULT_C;
  args->l = DEFAULT_FL;
  args->d = -1;
  args->m = DEFAULT_M;
  args->E = DEFAULT_E;
  args->L = DEFAULT_LL;
  args->R = DEFAULT_R;
  args->D = DEFAULT_D;
  args->P = 0;
  args->C = 0;
  args->S = 0;
  args->s = 0;
  args->p = 0;
  args->r = 0;
  args->a = 0;
  args->h = 0;
  args->v = 0;
  args->f = 0;

  while (1)
    {
      c = getopt_long (argc, argv, opt_string, long_options, &opti);
      if (c == -1)
	break;
      switch (c)
	{
	case 'c':		/* coverage */
	  args->c = atof (optarg);
	  break;
	case 'l':		/* mean fragment length */
	  args->l = atoi (optarg);
	  break;
	case 'd':		/* standard deviation of fragment length */
	  args->d = atoi (optarg);
	  break;
	case 'R':		/* mean read length */
	  args->R = atoi (optarg);
	  break;
	case 'D':		/* standard deviation of read length */
	  args->d = atoi (optarg);
	  break;
	case 'm':		/* minimum coverage of valid profile */
	  args->m = (float) atof (optarg);
	  break;
	case 'E':		/* sequencing error rate */
	  args->E = (float) atof (optarg);
	  break;
	case 'L':		/* maximum length of output lines */
	  args->L = (int) atoi (optarg);
	  break;
	case 's':		/* seed for random number generator */
	  args->s = (int) atoi (optarg);
	  if (args->s > 0)
	    args->s = -args->s;
	  if (args->s == 0)
	    args->s = -93213;
	  break;
	case 'S':		/* fixed start position */
	  args->S = (int) atoi (optarg);
	  break;
	case 'a':		/* alignment */
	  args->a = 1;
	  break;
	case 'p':		/* print profiles */
	  args->p = 1;
	  args->f = 1;		/* sample forward strand only */
	  args->a = 1;
	  break;
	case 'r':		/* curcular genome */
	  args->r = 1;
	  break;
	case 'P':		/* paired-end reads */
	  args->P = 1;
	  break;
	case 'C':               /* print single contig */
	  args->C = 1;
	  break;
	case 'h':		/* print help */
	  args->h = 1;
	  break;
	case 'v':		/* print version */
	  args->v = 1;
	  break;
	case '?':
	  break;
	default:
	  abort ();
	  return args;
	}
    }
  args->input_files = argv + optind;
  args->num_input_files = argc - optind;
  if(args->C && ! args->p)
    args->p = 1;
  if(args->d == -1)
    args->d = sqrt (args->l);
  return args;
}

void
print_usage ()
{
  printf ("Usage: sequencer [OPTION]... [FILE]...\n");
  printf ("Simulate shotgun sequencing.\n");
  printf ("Options:  \n");
  printf ("  -c --coverage=NUM          coverage; default: NUM=%3.1f\n",
	  DEFAULT_C);
  printf ("  -R --mean-read-length=NUM\n");
  printf ("                             default: NUM=%d\n", DEFAULT_R);
  printf ("  -D --sd-read-length=NUM\n");
  printf ("                             default: NUM=%4.2f\n",DEFAULT_D);
  printf ("  -P --paired-end-reads      default: single-end reads\n");
  printf ("  -l --mean-fragment-length=NUM\n");
  printf ("                             default: NUM=%d\n", DEFAULT_FL);
  printf ("  -d --sd-fragment-length=NUM\n");
  printf ("                             default: NUM=sqrt(mean-fragment-length)\n");
  printf
    ("  -m --minimum-coverage=NUM  minimum coverage of valid profile (only with -p)\n");
  printf ("                             default: NUM=%d\n", DEFAULT_M);
  printf ("  -E --error-rate=NUM        sequencing error rate\n");
  printf ("                             default: NUM=%4.2e\n", DEFAULT_E);
  printf ("  -L --output-line-length=NUM\n");
  printf ("                             default: %d\n", DEFAULT_LL);
  printf ("  -s --seed=NUM              seed for random number generator\n");
  printf
    ("                             default: random_seed.dat or system clock\n");
  printf ("  -S --start-position=NUM\n");
  printf
    ("                             default: random start position\n");
  printf ("  -p --print-profiles        default: print sequences of reads\n");
  printf ("  -C --single-contig         default: start new contig at every position not covered (implies -p)\n");
  printf ("  -a --alignment             implied by -p; default: no alignment\n");
  printf ("  -r --circular-genome       default: linear\n");
  printf
    ("  -h --help                  print this usage information and exit\n");
  printf
    ("  -v --version               print version information and exit\n");
}

void
print_version (char *version)
{
  printf ("sequencer %s\n", version);
  printf ("Written by Bernhard Haubold.\n");
  printf ("Distributed under the GNU General Public License.\n");
}
